/*
 * Copyright (c) 2003 Orion Hodson <oho@acm.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <signal.h>
#include <sysexits.h>

static int random_fetch_sizes = 0;
static int dummy_io = 0;

static uint64_t bytes_in = 0ULL;
static uint64_t bytes_out = 0ULL;

static void
dump_counters(void)
{
	fprintf(stderr, "bytes in %llu bytes out %llu\n", 
		bytes_in, bytes_out);
}

static void
sigint_handler(int sig)
{
	fprintf(stderr, "Caught SIGINT, exiting...\n");
	dump_counters();
	exit(sig);
}

static int
slurp(uint8_t* buf, uint32_t buf_bytes)
{
	size_t rs;

	if (random_fetch_sizes)
		buf_bytes = ((random() % (buf_bytes + 1)) / 4) * 4;

	bytes_in += buf_bytes;

	if (dummy_io)
		return buf_bytes;

	rs = fread(buf, 1, buf_bytes, stdin);
	return rs;
}

static void
pull_and_push(struct pcm_feeder* feeder)
{
	uint8_t b[256];
	size_t b_bytes, wrote;
	int fed;

	b_bytes = sizeof(b) / sizeof(b[0]); 
    
	for (;;) {
		fed = feed_rate(feeder, 0, b, b_bytes, 0);
		bytes_out += fed;
		if (fed == 0)
		    break;
		
		if (dummy_io)
			continue;

		wrote = fwrite(b, 1, fed, stdout);
		if (wrote != (size_t)fed) {
			fprintf(stderr, "Write error %d != %d\n", wrote, fed);
			perror("fwrite");
			break;
		}
	}
}

static void
usage(void)
{
	fprintf(stderr, "feed_rate [ -c <channels> ] <src_rate> <dst_rate>\n");
	exit(EX_USAGE);
}

int
main(int argc, char * const argv[])
{
	uint32_t srate, drate, channels;
	int ch;

	struct pcm_feeder feeder;

	channels = 1;
	signal(SIGINT, sigint_handler);
    
	while ((ch = getopt(argc, argv, "c:nr")) != -1) {
		switch (ch) {
		case 'c':
			channels = atoi(optarg);
			break;
		case 'n':
			dummy_io = 1;
			break;
		case 'r':
			random_fetch_sizes = 1;
			break;
		case '?':
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;
	    
	if (argc != 2)
		usage();

	srate = atoi(argv[0]);
	drate = atoi(argv[1]);

	if (channels != 2) {
		fprintf(stderr, 
			"Number of channels must be 2\n");
		usage();
	}

	if (srate > 48000 || srate == 0) {
		fprintf(stderr, 
			"Source sample rate must be in range (1-48000)\n");
		usage();
	}

	if (drate > 48000 || drate == 0) {
		fprintf(stderr, 
			"Destination sample rate must be in range (1-48000)\n");
		usage();
	}

	if (srate == 0 || drate == 0)
		usage();

	bzero(&feeder, sizeof(feeder));
    
	if (feed_rate_init(&feeder) < 0) {
		fprintf(stderr, "Could not initialize feeder\n");
		exit(EX_DATAERR);
	}

	if (feed_rate_set(&feeder, FEEDRATE_SRC, srate) < 0) {
		fprintf(stderr, "Could not set source rate\n");
		exit(EX_DATAERR);
	}

	if (feed_rate_set(&feeder, FEEDRATE_DST, drate) < 0) {
		fprintf(stderr, "Could not set destination rate\n");
		exit(EX_DATAERR);
	}

	pull_and_push(&feeder);

	feed_rate_get(&feeder, FEEDRATE_DST);

	feed_rate_free(&feeder);

	dump_counters();
    
	return 0;
}
