#include <inttypes.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#define OUT_SAMPLES 256

static void
mono_sin_gen(uint16_t amp, uint16_t trate, uint16_t srate)
{
	int i, p;
	int16_t s[OUT_SAMPLES];
	double tmp;

	i = 0; 
	p = 0;
	for (;;) {
		tmp = amp * sin(2 * M_PI * (i + p) * trate / (double)srate);
		s[i] = (int16_t)tmp;
		i++;
		if (i == OUT_SAMPLES) {
			fwrite(s, sizeof(s[0]), OUT_SAMPLES, stdout);
			fflush(stdout);
			p += OUT_SAMPLES;
			i = 0;
		}
	}
}

static void
stereo_sin_gen(uint16_t amp, uint16_t trate, uint16_t srate)
{
	int i, p;
	int16_t s[2 * OUT_SAMPLES];
	double tmp;

	i = 0; 
	p = 0;
	for (;;) {
		tmp = amp * sin(2 * M_PI * (i + p) * trate / (double)srate);
		s[2 * i] = (int16_t)tmp;
		s[2 * i + 1] = (int16_t)tmp;
		i++;
		if (i == OUT_SAMPLES) {
			fwrite(s, sizeof(s[0]), 2 * OUT_SAMPLES, stdout);
			fflush(stdout);
			p += OUT_SAMPLES;
			i = 0;
		}
	}
}

static void
mono_conv_gen(uint16_t amp, uint16_t trate, uint16_t srate)
{
	int i, p;
	int16_t s[OUT_SAMPLES];
	double tmp;

	i = 0; 
	p = 0;
	for (;;) {
		tmp = sin(2 * M_PI * (i + p) * trate * 0.05 / (double)srate);
		tmp *= tmp;
		tmp *= amp * sin(2 * M_PI * (i + p) * trate / (double)srate);

		s[i] = (int16_t)tmp;
		i++;
		if (i == OUT_SAMPLES) {
			fwrite(s, sizeof(s[0]), OUT_SAMPLES, stdout);
			fflush(stdout);
			p += OUT_SAMPLES;
			i = 0;
		}
	}
}

static void
stereo_conv_gen(uint16_t amp, uint16_t trate, uint16_t srate)
{
	int i, p;
	int16_t s[2 * OUT_SAMPLES];
	double tmp;

	i = 0; 
	p = 0;
	for (;;) {
		tmp = amp * sin(2 * M_PI * (i + p) * trate / (double)srate);
		tmp *= sin(2 * M_PI * (i + p) * trate * 0.1 / (double)srate);
		s[2 * i] = (int16_t)tmp;
		s[2 * i + 1] = (int16_t)tmp;
		i++;
		if (i == OUT_SAMPLES) {
			fwrite(s, sizeof(s[0]), 2 * OUT_SAMPLES, stdout);
			fflush(stdout);
			p += OUT_SAMPLES;
			i = 0;
		}
	}
}

static void
mono_sq_gen(uint16_t amp, uint16_t trate, uint16_t srate)
{
	int i, p;
	int16_t s[OUT_SAMPLES];

	int16_t d;
	int16_t tmp;

	d = srate / trate;
	i = 0; 
	p = 0;
	for (;;) {
		tmp = -amp + (((i + p) / d) % 2) * 2 * amp;
		s[i] = tmp;
		i++;
		if (i == OUT_SAMPLES) {
			fwrite(s, sizeof(s[0]), OUT_SAMPLES, stdout);
			fflush(stdout);
			p += OUT_SAMPLES;
			i = 0;
		}
	}
}

static void
stereo_sq_gen(uint16_t amp, uint16_t trate, uint16_t srate)
{
	int i, p;
	int16_t s[2 * OUT_SAMPLES];

	int16_t d;
	int16_t tmp;

	d = srate / trate;

	i = 0; 
	p = 0;
	for (;;) {
		tmp = -amp + (((i + p) / d) % 2) * 2 * amp;
		s[2 * i] = tmp;
		s[2 * i + 1] = tmp;
		i++;
		if (i == OUT_SAMPLES) {
			fwrite(s, sizeof(s[0]), 2 * OUT_SAMPLES, stdout);
			fflush(stdout);
			p += OUT_SAMPLES;
			i = 0;
		}
	}
}

static void
mono_saw_gen(uint16_t amp, uint16_t trate, uint16_t srate)
{
	int i, p;
	int16_t s[OUT_SAMPLES];

	int16_t d;
	int16_t tmp;

	d = srate / trate;
	i = 0; 
	p = 0;
	for (;;) {
		tmp = -amp + ((i + p) % d) * 2 * amp / d;
		s[i] = tmp;
		i++;
		if (i == OUT_SAMPLES) {
			fwrite(s, sizeof(s[0]), OUT_SAMPLES, stdout);
			fflush(stdout);
			p += OUT_SAMPLES;
			i = 0;
		}
	}
}

static void
stereo_saw_gen(uint16_t amp, uint16_t trate, uint16_t srate)
{
	int i, p;
	int16_t s[2 * OUT_SAMPLES];

	int16_t d;
	int16_t tmp;

	d = srate / trate;

	i = 0; 
	p = 0;
	for (;;) {
		tmp = -amp + ((i + p) % d) * 2 * amp / d;
		s[2 * i] = tmp;
		s[2 * i + 1] = tmp;
		i++;
		if (i == OUT_SAMPLES) {
			fwrite(s, sizeof(s[0]), 2 * OUT_SAMPLES, stdout);
			fflush(stdout);
			p += OUT_SAMPLES;
			i = 0;
		}
	}
}

struct gen {
	const char* name;
	void (*spew[2])(uint16_t, uint16_t, uint16_t);
} generators[] = {
	{ "si", { mono_sin_gen, stereo_sin_gen } },
	{ "conv", { mono_conv_gen, stereo_conv_gen } },
	{ "sq", { mono_sq_gen, stereo_sq_gen } },
	{ "saw", { mono_saw_gen, stereo_saw_gen } },
};
#define ngenerators (sizeof(generators) / sizeof(generators[0]))

static int
name_to_generator(const char* name)
{
	uint32_t i;
	for (i = 0; i < ngenerators; i++) {
		if (0 == strncasecmp(name, generators[i].name, 2)) {
			return i;
		}
	}
	return 0;
}

static void
usage(void)
{
	fprintf(stderr, 
		"tonegen [ -c <channels> ] [ -a <amplitute> ] [ -s sq|sin ]"
		"<tone_rate> <smpl_rate>");
	exit(EX_USAGE);
}

int main(int argc, char * const * argv)
{
	int channels, amp, trate, srate, shape, ch;

	shape = 0;
	channels = 1;
	amp = 32767;

	while ((ch = getopt(argc, argv, "a:c:s:")) != -1) {
                switch (ch) {
		case 'a':
			amp = atoi(optarg);
			break;
                case 'c':
                        channels = atoi(optarg);
                        break;
		case 's':
			shape = name_to_generator(optarg);
			break;
                case '?':
                default:
                        usage();
                }
        }
	argc -= optind;
        argv += optind;
            
        if (argc != 2)
                usage();

	trate = atoi(argv[0]);
	srate = atoi(argv[1]);

	if (channels < 1 || channels > 2) {
		fprintf(stderr, "Channels must be in range 1--2\n");
		usage();
	}

	if (amp < 1 || amp > 32767) {
		fprintf(stderr, "Amplitude must be in range 1--32767\n");
		usage();
	}

	if (srate == 0 || srate > 48000) {
		fprintf(stderr, "smpl_rate must be in range 1--48000");
		usage();
	}

	if (trate == 0 || trate >= srate / 2) {
		fprintf(stderr, "tone_rate must be in range 1--smpl_rate / 2");
		usage();
	}

	generators[shape].spew[channels - 1](amp, trate, srate);

	return 0;
}
